# -*- coding: utf-8 -*-

"""
Amazon Augmented AI for Human.
"""


from ._version import __version__

__short_description__ = "Amazon Augmented AI for Human."
__license__ = "MIT"
__author__ = "Sanhe Hu"
__author_email__ = "husanhe@gmail.com"
__github_username__ = "MacHu-GWU"


try:
    from .better_boto.human_task_ui import (
        get_task_template_arn,
        get_task_template_console_url,
        parse_task_template_name_from_arn,
        is_hil_task_template_exists,
        create_human_task_ui,
        delete_human_task_ui,
        deploy_hil_task_template,
        remove_hil_task_template,
    )
    from .better_boto.flow_definition import (
        FlowDefinitionStatusEnum,
        FlowDefinition,
        get_flow_definition_arn,
        get_flow_definition_console_url,
        parse_flow_definition_name_from_arn,
        is_flow_definition_exists,
        create_flow_definition,
        delete_flow_definition,
        remove_flow_definition,
        deploy_flow_definition,
    )
    from .better_boto.human_loop import (
        HumanLoopStatusEnum,
        HumanLoop,
        parse_team_name_from_private_team_arn,
        get_workspace_signin_url,
        get_hil_console_url,
        parse_hil_name_from_hil_arn,
        describe_human_loop,
        start_human_loop,
        get_human_loop_details,
        stop_human_loop,
        delete_human_loop,
        list_human_loops,
    )
    from .tools import (
        render_task_template,
    )
    from .tagging import (
        to_tag_list,
        to_tag_dict,
    )
    from .helper import (
        sha256_of_bytes,
        vprint,
    )
except ImportError as e:  # pragma: no cover
    print(e)



# -*- coding: utf-8 -*-

import typing as T


def to_tag_list(tags: T.Dict[str, str]) -> T.List[T.Dict[str, str]]:
    return [dict(Key=k, Value=v) for k, v in tags.items()]


def to_tag_dict(tags: T.List[T.Dict[str, str]]) -> T.Dict[str, str]:
    return {dct["Key"]: dct["Value"] for dct in tags}



# -*- coding: utf-8 -*-

import hashlib


def sha256_of_bytes(b: bytes) -> str:
    m = hashlib.sha256()
    m.update(b)
    return m.hexdigest()


def vprint(msg: str, verbose: bool):
    if verbose:
        print(msg)



# -*- coding: utf-8 -*-

import typing as T

from light_emoji import common as emojis
from boto_session_manager import BotoSesManager

from ..helper import sha256_of_bytes, vprint
from ..tagging import to_tag_list


# --- Low level API
def get_task_template_arn(
    aws_account_id: str,
    aws_region: str,
    task_template_name: str,
) -> str:
    return (
        f"arn:aws:sagemaker:{aws_region}:{aws_account_id}:human-task-ui"
        f"/{task_template_name}"
    )


def get_task_template_console_url(
    aws_region: str,
    task_template_name: str,
) -> str:
    return (
        f"https://{aws_region}.console.aws.amazon.com/sagemaker"
        f"/groundtruth?region={aws_region}#/a2i/worker-task-templates"
        f"/{task_template_name}"
    )


def parse_task_template_name_from_arn(arn: str) -> str:
    """
    Example:

        >>> parse_task_template_name_from_arn("arn:aws:sagemaker:us-east-1:111122223333:human-task-ui/my-task-template")
        'my-task-template'
    """
    return arn.split("/")[-1]


def is_hil_task_template_exists(
    bsm: BotoSesManager,
    task_template_name: str,
) -> T.Tuple[bool, dict]:
    """
    Reference:

    - https://boto3.amazonaws.com/v1/documentation/api/latest/reference/services/sagemaker.html#SageMaker.Client.describe_human_task_ui

    :return: tuple of two item, first item is a boolean value, second value is
        the response of ``describe_human_task_ui()``, you can call it task ui details.
    """
    try:
        response = bsm.sagemaker_client.describe_human_task_ui(
            HumanTaskUiName=task_template_name,
        )
        return True, response
    except Exception as e:
        if "does not exist" in str(e):
            return False, {}
        else:
            raise e


def create_human_task_ui(
    bsm: BotoSesManager,
    task_template_name: str,
    task_template_content: str,
    tags: T.Optional[T.Dict[str, str]] = None,
) -> dict:
    """
    ref: https://boto3.amazonaws.com/v1/documentation/api/latest/reference/services/sagemaker.html#SageMaker.Client.create_human_task_ui
    """
    kwargs = dict(
        HumanTaskUiName=task_template_name,
        UiTemplate={"Content": task_template_content},
    )
    if tags:
        kwargs["Tags"] = to_tag_list(tags)
    return bsm.sagemaker_client.create_human_task_ui(**kwargs)


def delete_human_task_ui(
    bsm: BotoSesManager,
    task_template_name: str,
):
    """
    Reference:

    - https://boto3.amazonaws.com/v1/documentation/api/latest/reference/services/sagemaker.html#SageMaker.Client.delete_human_task_ui
    """
    return bsm.sagemaker_client.delete_human_task_ui(HumanTaskUiName=task_template_name)


# --- High level API
def deploy_hil_task_template(
    bsm: BotoSesManager,
    task_template_name: str,
    task_template_content: str,
    tags: T.Optional[T.Dict[str, str]] = None,
    verbose: bool = True,
) -> T.Optional[dict]:
    """
    Deploy HIL task template, smartly.

    :return: if deployment happens, then return the response of ``create_human_task_ui()``,
        otherwise return ``None``.
    """
    vprint(f"{emojis.deploy} Deploy Human in Loop task template", verbose)
    task_template_console_url = get_task_template_console_url(
        aws_region=bsm.aws_region,
        task_template_name=task_template_name,
    )
    vprint(f"  preview at {task_template_console_url}", verbose)
    flag, response = is_hil_task_template_exists(
        bsm=bsm,
        task_template_name=task_template_name,
    )
    if flag:
        content_sha256 = response["UiTemplate"]["ContentSha256"]
        if content_sha256 == sha256_of_bytes(task_template_content.encode("utf-8")):
            vprint(
                "  a HIL task template with the same content already exists, do nothing.",
                verbose,
            )
            return None
        else:
            delete_human_task_ui(
                bsm=bsm,
                task_template_name=task_template_name,
            )
            response = create_human_task_ui(
                bsm=bsm,
                task_template_name=task_template_name,
                task_template_content=task_template_content,
                tags=tags,
            )
    else:
        response = create_human_task_ui(
            bsm, task_template_name, task_template_content, tags
        )
    vprint(
        f"  {emojis.succeeded} Successfully deployed task ui template {task_template_name!r}",
        verbose,
    )
    return response


def remove_hil_task_template(
    bsm: BotoSesManager,
    task_template_name: str,
    verbose: bool = True,
) -> bool:
    """
    Remove HIL task template, smartly.

    :return: a boolean flag indicating whether the delete operation happens.
    """
    vprint(f"{emojis.delete} Remove Human in Loop task template", verbose)
    task_template_console_url = get_task_template_console_url(
        aws_region=bsm.aws_region,
        task_template_name=task_template_name,
    )
    vprint(f"  verify at {task_template_console_url}", verbose)
    flag, response = is_hil_task_template_exists(
        bsm=bsm,
        task_template_name=task_template_name,
    )
    if flag:
        delete_human_task_ui(
            bsm=bsm,
            task_template_name=task_template_name,
        )
    else:
        vprint("  HIL task template doesn't exists, do nothing.", verbose)

    vprint(
        f"  {emojis.succeeded} Successfully removed task ui template {task_template_name!r}",
        verbose,
    )
    return not flag



# -*- coding: utf-8 -*-

import typing as T
import enum
import dataclasses
from datetime import datetime

from light_emoji import common as emojis
from boto_session_manager import BotoSesManager

from ..tagging import to_tag_list
from ..helper import vprint
from ..waiter import Waiter

from .human_task_ui import (
    get_task_template_arn,
)


# --- Data Model
@dataclasses.dataclass
class HumanLoopConfig:
    """
    An attribute of :class:`FlowDefinition`.
    """

    work_team_arn: T.Optional[str] = dataclasses.field(default=None)
    human_task_ui_arn: T.Optional[str] = dataclasses.field(default=None)
    task_title: T.Optional[str] = dataclasses.field(default=None)
    task_description: T.Optional[str] = dataclasses.field(default=None)
    task_count: T.Optional[int] = dataclasses.field(default=None)
    task_availability_lifetime_in_seconds: T.Optional[int] = dataclasses.field(
        default=None
    )
    task_time_limit_in_seconds: T.Optional[int] = dataclasses.field(default=None)

    @property
    def human_task_ui_name(self) -> str:
        return self.task_title


@dataclasses.dataclass
class OutputConfig:
    """
    An attribute of :class:`FlowDefinition`.
    """

    s3_output_path: T.Optional[str] = dataclasses.field(default=None)
    kms_key_id: T.Optional[str] = dataclasses.field(default=None)


class FlowDefinitionStatusEnum(str, enum.Enum):
    """
    human review workflow definition status enumeration.
    """

    Initializing = "Initializing"
    Active = "Active"
    Failed = "Failed"
    Deleting = "Deleting"


@dataclasses.dataclass
class FlowDefinition:
    """
    The data model of a human review workflow definition.

    :param data: the raw data from the ``describe_flow_definition`` api response.
    """

    arn: T.Optional[str] = dataclasses.field(default=None)
    name: T.Optional[str] = dataclasses.field(default=None)
    status: T.Optional[str] = dataclasses.field(default=None)
    creation_time: T.Optional[datetime] = dataclasses.field(default=None)
    role_arn: T.Optional[str] = dataclasses.field(default=None)
    human_loop_config: T.Optional[HumanLoopConfig] = dataclasses.field(default=None)
    output_config: T.Optional[OutputConfig] = dataclasses.field(default=None)
    data: T.Optional[dict] = dataclasses.field(default=None)

    @classmethod
    def from_describe_flow_definition_response(cls, response: dict) -> "FlowDefinition":
        """
        Reference:

        - https://boto3.amazonaws.com/v1/documentation/api/latest/reference/services/sagemaker/client/describe_flow_definition.html
        """
        return cls(
            arn=response["FlowDefinitionArn"],
            name=response["FlowDefinitionName"],
            status=response["FlowDefinitionStatus"],
            role_arn=response["RoleArn"],
            creation_time=response["CreationTime"],
            human_loop_config=HumanLoopConfig(
                work_team_arn=response["HumanLoopConfig"]["WorkteamArn"],
                human_task_ui_arn=response["HumanLoopConfig"]["HumanTaskUiArn"],
                task_title=response["HumanLoopConfig"]["TaskTitle"],
                task_description=response["HumanLoopConfig"]["TaskDescription"],
                task_count=response["HumanLoopConfig"]["TaskCount"],
                task_availability_lifetime_in_seconds=response["HumanLoopConfig"].get(
                    "TaskAvailabilityLifetimeInSeconds"
                ),
                task_time_limit_in_seconds=response["HumanLoopConfig"].get(
                    "TaskTimeLimitInSeconds"
                ),
            ),
            output_config=OutputConfig(
                s3_output_path=response["OutputConfig"]["S3OutputPath"],
                kms_key_id=response["OutputConfig"].get("KmsKeyId"),
            ),
            data=response,
        )

    def is_initializing(self) -> bool:
        return self.status == FlowDefinitionStatusEnum.Initializing.value

    def is_active(self) -> bool:
        return self.status == FlowDefinitionStatusEnum.Active.value

    def is_failed(self) -> bool:
        return self.status == FlowDefinitionStatusEnum.Failed.value

    def is_deleting(self) -> bool:
        return self.status == FlowDefinitionStatusEnum.Deleting.value


# --- Low level API
def get_flow_definition_arn(
    aws_account_id: str,
    aws_region: str,
    flow_definition_name: str,
) -> str:
    return (
        f"arn:aws:sagemaker:{aws_region}:{aws_account_id}:flow-definition"
        f"/{flow_definition_name}"
    )


def get_flow_definition_console_url(
    aws_region: str,
    flow_definition_name: str,
) -> str:
    return (
        f"https://{aws_region}.console.aws.amazon.com/sagemaker"
        f"/groundtruth?region={aws_region}#"
        f"/a2i/human-review-workflows/{flow_definition_name}"
    )


def parse_flow_definition_name_from_arn(arn: str) -> str:
    """
    Example:

        >>> parse_flow_definition_name_from_arn("arn:aws:sagemaker:us-east-1:111122223333:flow-definition/my-flow")
        'my-flow'
    """
    return arn.split("/")[-1]


def create_flow_definition(
    bsm: BotoSesManager,
    flow_definition_name: str,
    flow_execution_role_arn: str,
    labeling_team_arn: str,
    output_bucket: str,
    output_key: str,
    task_template_name: str,
    task_description: str,
    task_count: int,
    task_availability_life_time_in_seconds: T.Optional[int] = None,
    task_time_limit_in_seconds: T.Optional[int] = None,
    tags: T.Optional[T.Dict[str, str]] = None,
) -> dict:
    """
    Reference:

    - https://boto3.amazonaws.com/v1/documentation/api/latest/reference/services/sagemaker.html#SageMaker.Client.create_flow_definition
    """
    task_template_arn = get_task_template_arn(
        aws_account_id=bsm.aws_account_id,
        aws_region=bsm.aws_region,
        task_template_name=task_template_name,
    )
    if output_key.endswith("/"):  # aws will treat this as a folder
        output_key = output_key[:-1]
    human_loop_config = {
        "WorkteamArn": labeling_team_arn,
        "HumanTaskUiArn": task_template_arn,
        "TaskTitle": task_template_name,
        "TaskDescription": task_description,
        "TaskCount": task_count,
    }
    if task_availability_life_time_in_seconds is not None:
        human_loop_config[
            "TaskAvailabilityLifetimeInSeconds"
        ] = task_availability_life_time_in_seconds
    if task_time_limit_in_seconds is not None:
        human_loop_config["TaskTimeLimitInSeconds"] = task_time_limit_in_seconds
    kwargs = dict(
        FlowDefinitionName=flow_definition_name,
        HumanLoopConfig=human_loop_config,
        OutputConfig={
            "S3OutputPath": f"s3://{output_bucket}/{output_key}",
        },
        RoleArn=flow_execution_role_arn,
    )
    if tags:
        kwargs["Tags"] = to_tag_list(tags)
    response = bsm.sagemaker_client.create_flow_definition(**kwargs)
    return response


def delete_flow_definition(
    bsm: BotoSesManager,
    flow_definition_name: str,
):
    """
    Reference:

    - https://boto3.amazonaws.com/v1/documentation/api/latest/reference/services/sagemaker.html#SageMaker.Client.delete_flow_definition
    """
    return bsm.sagemaker_client.delete_flow_definition(
        FlowDefinitionName=flow_definition_name
    )


# --- High level API
def is_flow_definition_exists(
    bsm: BotoSesManager,
    flow_definition_name: str,
) -> T.Tuple[bool, T.Optional[FlowDefinition]]:
    """
    Reference:

    - https://boto3.amazonaws.com/v1/documentation/api/latest/reference/services/sagemaker.html#SageMaker.Client.describe_flow_definition

    :return: tuple of two item, first item is a boolean value, second value is
        the response of ``describe_flow_definition()``, you can call it flow details.
    """
    try:
        response = bsm.sagemaker_client.describe_flow_definition(
            FlowDefinitionName=flow_definition_name
        )
        return True, FlowDefinition.from_describe_flow_definition_response(response)
    except Exception as e:
        if "does not exist" in str(e):
            return False, None
        else:
            raise e


def remove_flow_definition(
    bsm: BotoSesManager,
    flow_definition_name: str,
    wait: bool = True,
    wait_timeout: int = 30,
    verbose: bool = True,
):
    """
    High level api to remove a human review workflow definition.
    """
    vprint(
        f"{emojis.delete} Remove Human review workflow definition, it may takes 30 sec ~ 1 minute",
        verbose,
    )
    flow_definition_console_url = get_flow_definition_console_url(
        bsm.aws_region,
        flow_definition_name,
    )
    vprint(f"  Preview at {flow_definition_console_url}", verbose)

    is_flow_exists, flow_def = is_flow_definition_exists(
        bsm=bsm,
        flow_definition_name=flow_definition_name,
    )
    if is_flow_exists:
        delete_flow_definition(
            bsm=bsm,
            flow_definition_name=flow_definition_name,
        )
        if wait:
            for _ in Waiter(delays=1, timeout=wait_timeout, indent=2, verbose=verbose):
                is_flow_exists, flow_def = is_flow_definition_exists(
                    bsm=bsm,
                    flow_definition_name=flow_definition_name,
                )
                if is_flow_exists is False:
                    vprint("", verbose)
                    break
                if flow_def.status == "Failed":
                    raise Exception("Failed!")
    else:
        vprint("  Flow definition doesn't exists, do nothing.", verbose)
    vprint(
        f"  {emojis.succeeded} Successfully delete flow definition {flow_definition_name!r}",
        verbose,
    )


def deploy_flow_definition(
    bsm: BotoSesManager,
    flow_definition_name: str,
    flow_execution_role_arn: str,
    labeling_team_arn: str,
    output_bucket: str,
    output_key: str,
    task_template_name: str,
    task_description: str,
    task_count: int,
    task_availability_life_time_in_seconds: T.Optional[int] = None,
    task_time_limit_in_seconds: T.Optional[int] = None,
    tags: T.Optional[T.Dict[str, str]] = None,
    wait: bool = True,
    wait_delay: int = 3,
    wait_timeout: int = 30,
    verbose: bool = True,
) -> T.Optional[dict]:
    """
    High level api to deploy a Human review workflow, smartly.

    :return: if deployment happens, then return the response of ``create_flow_definition()``,
        otherwise return ``None``.
    """
    vprint(
        f"{emojis.deploy} Deploy Human review workflow definition, it may takes 30 sec ~ 1 minute",
        verbose,
    )
    flow_definition_console_url = get_flow_definition_console_url(
        aws_region=bsm.aws_region,
        flow_definition_name=flow_definition_name,
    )
    vprint(f"  preview at {flow_definition_console_url}", verbose)
    is_flow_exists, flow_def = is_flow_definition_exists(
        bsm=bsm,
        flow_definition_name=flow_definition_name,
    )
    if is_flow_exists:
        no_change_flag = (
            (flow_execution_role_arn == flow_def.role_arn)
            and (task_template_name == flow_def.human_loop_config.task_title)
            and (task_description == flow_def.human_loop_config.task_description)
            and (task_count == flow_def.human_loop_config.task_count)
        )
        if task_availability_life_time_in_seconds is not None:
            no_change_flag = no_change_flag and (
                task_availability_life_time_in_seconds
                == flow_def.human_loop_config.task_availability_lifetime_in_seconds
            )
        if task_time_limit_in_seconds is not None:
            no_change_flag = no_change_flag and (
                task_time_limit_in_seconds
                == flow_def.human_loop_config.task_time_limit_in_seconds
            )
        if output_key.endswith("/"):
            output_key = output_key[:-1]
        no_change_flag = no_change_flag and (
            (
                flow_def.output_config.s3_output_path
                == f"s3://{output_bucket}/{output_key}"
            )
        )
        # no need to deploy
        if no_change_flag:
            vprint("  No configuration changed, do nothing.", verbose)
            return None
        # remove existing one first
        remove_flow_definition(
            bsm=bsm,
            flow_definition_name=flow_definition_name,
            wait=True,
            verbose=verbose,
        )

    vprint("Create Human review workflow definition ...", verbose)
    response = create_flow_definition(
        bsm,
        flow_definition_name=flow_definition_name,
        flow_execution_role_arn=flow_execution_role_arn,
        labeling_team_arn=labeling_team_arn,
        output_bucket=output_bucket,
        output_key=output_key,
        task_template_name=task_template_name,
        task_description=task_description,
        task_count=task_count,
        task_availability_life_time_in_seconds=task_availability_life_time_in_seconds,
        task_time_limit_in_seconds=task_time_limit_in_seconds,
        tags=tags,
    )

    if wait:
        for _ in Waiter(
            delays=wait_delay, timeout=wait_timeout, indent=2, verbose=verbose
        ):
            is_flow_exists, flow_def = is_flow_definition_exists(
                bsm=bsm,
                flow_definition_name=flow_definition_name,
            )
            # create_flow_definition() may return before the flow definition
            # is created
            if is_flow_exists is False:
                continue

            if flow_def.is_active():
                vprint("", verbose)
                break

            if flow_def.is_failed():
                raise Exception(
                    f"Creating Human Review Workflow {flow_definition_name!r} Failed"
                )

    vprint(
        f"  {emojis.succeeded} Successfully deployed flow definition {flow_definition_name!r}",
        verbose,
    )
    return response



# -*- coding: utf-8 -*-

import pytest
from pytest import raises, approx


def test():
    from aws_a2i import api

    _ = api.get_task_template_arn
    _ = api.get_task_template_console_url
    _ = api.parse_task_template_name_from_arn
    _ = api.is_hil_task_template_exists
    _ = api.create_human_task_ui
    _ = api.delete_human_task_ui
    _ = api.deploy_hil_task_template
    _ = api.remove_hil_task_template
    _ = api.FlowDefinitionStatusEnum
    _ = api.FlowDefinition
    _ = api.get_flow_definition_arn
    _ = api.get_flow_definition_console_url
    _ = api.parse_flow_definition_name_from_arn
    _ = api.is_flow_definition_exists
    _ = api.create_flow_definition
    _ = api.delete_flow_definition
    _ = api.remove_flow_definition
    _ = api.deploy_flow_definition
    _ = api.HumanAnswer
    _ = api.HumanLoopOutput
    _ = api.HumanLoopStatusEnum
    _ = api.HumanLoop
    _ = api.parse_team_name_from_private_team_arn
    _ = api.get_workspace_signin_url
    _ = api.get_hil_console_url
    _ = api.parse_hil_name_from_hil_arn
    _ = api.describe_human_loop
    _ = api.start_human_loop
    _ = api.get_human_loop_details
    _ = api.stop_human_loop
    _ = api.delete_human_loop
    _ = api.list_human_loops
    _ = api.render_task_template
    _ = api.to_tag_list
    _ = api.to_tag_dict
    _ = api.sha256_of_bytes
    _ = api.vprint
    _ = api.Play


if __name__ == "__main__":
    import os

    basename = os.path.basename(__file__)
    pytest.main([basename, "-s", "--tb=native"])



# -*- coding: utf-8 -*-

import typing as T
import sys
import subprocess
from pathlib import Path

try:
    from box import Box
    from liquid import Template

    HAS_DEPENDENCIES = True
except ImportError as e:
    HAS_DEPENDENCIES = False


def _warn_if_no_dependencies():
    if HAS_DEPENDENCIES is False:
        raise ImportError(f"cannot import 'box' or 'liquid' package")


def render_task_template(
    task_template_content: str,
    input_data: dict,
    path_task_ui_html: T.Union[str, Path, T.Any],
    preview: bool = True,
):
    _warn_if_no_dependencies()

    # read liquid template
    template = Template(task_template_content)

    # convert task data to box, so it support dot notation
    task = Box({"input": input_data})

    # render template
    content = template.render(task=task)

    # write template to html file
    path_task_ui_html = Path(path_task_ui_html)
    path_task_ui_html.write_text(content)

    # open html in browser
    if preview:
        if sys.platform in ["win32", "cygwin"]:
            open_command = "start"
        else:
            open_command = "open"
        subprocess.run([open_command, str(path_task_ui_html)])



# -*- coding: utf-8 -*-

from .better_boto.human_task_ui import get_task_template_arn
from .better_boto.human_task_ui import get_task_template_console_url
from .better_boto.human_task_ui import parse_task_template_name_from_arn
from .better_boto.human_task_ui import is_hil_task_template_exists
from .better_boto.human_task_ui import create_human_task_ui
from .better_boto.human_task_ui import delete_human_task_ui
from .better_boto.human_task_ui import deploy_hil_task_template
from .better_boto.human_task_ui import remove_hil_task_template
from .better_boto.flow_definition import FlowDefinitionStatusEnum
from .better_boto.flow_definition import FlowDefinition
from .better_boto.flow_definition import get_flow_definition_arn
from .better_boto.flow_definition import get_flow_definition_console_url
from .better_boto.flow_definition import parse_flow_definition_name_from_arn
from .better_boto.flow_definition import is_flow_definition_exists
from .better_boto.flow_definition import create_flow_definition
from .better_boto.flow_definition import delete_flow_definition
from .better_boto.flow_definition import remove_flow_definition
from .better_boto.flow_definition import deploy_flow_definition
from .better_boto.human_loop import HumanAnswer
from .better_boto.human_loop import HumanLoopOutput
from .better_boto.human_loop import HumanLoopStatusEnum
from .better_boto.human_loop import HumanLoop
from .better_boto.human_loop import parse_team_name_from_private_team_arn
from .better_boto.human_loop import get_workspace_signin_url
from .better_boto.human_loop import get_hil_console_url
from .better_boto.human_loop import parse_hil_name_from_hil_arn
from .better_boto.human_loop import describe_human_loop
from .better_boto.human_loop import start_human_loop
from .better_boto.human_loop import get_human_loop_details
from .better_boto.human_loop import stop_human_loop
from .better_boto.human_loop import delete_human_loop
from .better_boto.human_loop import list_human_loops
from .tools import render_task_template
from .tagging import to_tag_list
from .tagging import to_tag_dict
from .helper import sha256_of_bytes
from .helper import vprint
from .playground import Play


# -*- coding: utf-8 -*-

import typing as T
import json
import enum
import dataclasses
from datetime import datetime, timezone

from light_emoji import common as emojis
from iterproxy import IterProxy
from boto_session_manager import BotoSesManager

from ..helper import vprint

from .flow_definition import (
    parse_flow_definition_name_from_arn,
)

# --- Data Model
date_fmt = "%Y-%m-%dT%H:%M:%S.%fZ"


@dataclasses.dataclass
class HumanAnswer:
    acceptanceTime: str
    answerContent: dict
    submissionTime: str
    timeSpentInSeconds: float
    workerId: str
    workerMetadata: dict

    @classmethod
    def from_dict(cls, dct: dict):
        return cls(**dct)

    @property
    def acceptance_datetime(self) -> datetime:
        return datetime.strptime(self.acceptanceTime, date_fmt).replace(
            tzinfo=timezone.utc
        )

    @property
    def submission_datetime(self) -> datetime:
        return datetime.strptime(self.submissionTime, date_fmt).replace(
            tzinfo=timezone.utc
        )


@dataclasses.dataclass
class HumanLoopOutput:
    """
    Human loop output data model.
    """

    flowDefinitionArn: str
    humanAnswers: T.List[HumanAnswer]
    humanLoopName: str
    inputContent: dict

    @classmethod
    def from_dict(cls, dct: dict):
        dct["humanAnswers"] = [
            HumanAnswer.from_dict(dct) for dct in dct.get("humanAnswers", [])
        ]
        return cls(**dct)


class HumanLoopStatusEnum(str, enum.Enum):
    """
    Human loop status enumeration
    """

    InProgress = "InProgress"
    Failed = "Failed"
    Completed = "Completed"
    Stopped = "Stopped"
    Stopping = "Stopping"


@dataclasses.dataclass
class HumanLoop:
    """
    The data model of a human loop.

    :param data: the raw data from the ``describe_human_loop`` api response.
    """

    creation_time: T.Optional[datetime] = dataclasses.field(default=None)
    failure_reason: T.Optional[str] = dataclasses.field(default=None)
    failure_code: T.Optional[str] = dataclasses.field(default=None)
    human_loop_status: T.Optional[str] = dataclasses.field(default=None)
    human_loop_name: T.Optional[str] = dataclasses.field(default=None)
    human_loop_arn: T.Optional[str] = dataclasses.field(default=None)
    flow_definition_arn: T.Optional[str] = dataclasses.field(default=None)
    human_loop_output_s3uri: T.Optional[str] = dataclasses.field(default=None)
    data: T.Optional[dict] = dataclasses.field(default=None)

    def get_details(self, bsm: BotoSesManager) -> "HumanLoop":
        """
        Call ``describe_human_loop`` API to fetch additional details.
        """
        if self.human_loop_name is not None:
            human_loop = get_human_loop_details(
                bsm=bsm,
                human_loop_name_or_arn=self.human_loop_name,
            )
        elif self.human_loop_arn is not None:
            human_loop = get_human_loop_details(
                bsm=bsm,
                human_loop_name_or_arn=self.human_loop_arn,
            )
        else:
            raise NotImplementedError
        self.creation_time = human_loop.creation_time
        self.failure_reason = human_loop.failure_reason
        self.failure_code = human_loop.failure_code
        self.human_loop_status = human_loop.human_loop_status
        self.human_loop_name = human_loop.human_loop_name
        self.human_loop_arn = human_loop.human_loop_arn
        self.flow_definition_arn = human_loop.flow_definition_arn
        self.human_loop_output_s3uri = human_loop.human_loop_output_s3uri
        self.data = human_loop.data
        return self

    def is_in_progress(self) -> bool:
        return self.human_loop_status == HumanLoopStatusEnum.InProgress.value

    def is_failed(self) -> bool:
        return self.human_loop_status == HumanLoopStatusEnum.Failed.value

    def is_completed(self) -> bool:
        return self.human_loop_status == HumanLoopStatusEnum.Completed.value

    def is_stopped(self) -> bool:
        return self.human_loop_status == HumanLoopStatusEnum.Stopped.value

    def is_stopping(self) -> bool:
        return self.human_loop_status == HumanLoopStatusEnum.Stopping.value

    def get_output(self, bsm: BotoSesManager) -> HumanLoopOutput:
        parts = self.human_loop_output_s3uri.split("/", 3)
        bucket = parts[2]
        key = parts[3]
        res = bsm.s3_client.get_object(
            Bucket=bucket,
            Key=key,
        )
        data = json.loads(res["Body"].read())
        return HumanLoopOutput.from_dict(data)


# --- Low level API
def parse_team_name_from_private_team_arn(arn: str) -> str:
    """
    Example:

         >>> parse_team_name_from_private_team_arn(
            "arn:aws:sagemaker:us-east-1:111122223333:workteam/private-crowd/my-workforce"
         )
         'my-workforce'
    """
    return arn.split("/")[-1]


def get_workspace_signin_url(
    bsm: BotoSesManager,
    work_team_name: str,
) -> str:
    """
    Example:

        >>> get_workspace_signin_url(bsm, "my-workforce")
        'https://1a2b3c4d5e.labeling.us-east-1.sagemaker.aws'
    """
    response = bsm.sagemaker_client.describe_workteam(WorkteamName=work_team_name)
    sub_domain = response["Workteam"]["SubDomain"]
    return "https://" + sub_domain


def get_hil_console_url(
    aws_region: str,
    flow_definition_name: str,
    hil_name: str,
) -> str:
    return (
        f"https://{aws_region}.console.aws.amazon.com/sagemaker"
        f"/groundtruth?region={aws_region}#/a2i/human-review-workflows"
        f"/{flow_definition_name}/human-loops/{hil_name}"
    )


def parse_hil_name_from_hil_arn(arn: str) -> str:
    """
    Example:

        >>> parse_hil_name_from_hil_arn("arn:aws:sagemaker:us-east-1:111122223333:human-loop/4d7b0711-0e9e-48c5-9df9-a03692cdcd8b")
        '4d7b0711-0e9e-48c5-9df9-a03692cdcd8b'
    """
    return arn.split("/")[-1]


def describe_human_loop(
    bsm: BotoSesManager,
    human_loop_name: str,
) -> dict:
    """
    Reference:

    - https://boto3.amazonaws.com/v1/documentation/api/latest/reference/services/sagemaker-a2i-runtime.html#AugmentedAIRuntime.Client.describe_human_loop
    """
    return bsm.sagemaker_a2i_runtime_client.describe_human_loop(
        HumanLoopName=human_loop_name,
    )


def stop_human_loop(
    bsm: BotoSesManager,
    human_loop_name: str,
):
    """
    Reference:

    - https://boto3.amazonaws.com/v1/documentation/api/latest/reference/services/sagemaker-a2i-runtime.html#AugmentedAIRuntime.Client.stop_human_loop
    """
    return bsm.sagemaker_a2i_runtime_client.stop_human_loop(
        HumanLoopName=human_loop_name,
    )


def delete_human_loop(
    bsm: BotoSesManager,
    human_loop_name: str,
):
    """
    Reference:

    - https://boto3.amazonaws.com/v1/documentation/api/latest/reference/services/sagemaker-a2i-runtime.html#AugmentedAIRuntime.Client.delete_human_loop
    """
    return bsm.sagemaker_a2i_runtime_client.delete_human_loop(
        HumanLoopName=human_loop_name,
    )


# --- High level API
def start_human_loop(
    bsm: BotoSesManager,
    human_loop_name: str,
    flow_definition_arn: str,
    input_data: dict,
    data_attributes: T.Optional[dict] = None,
    verbose: bool = True,
) -> str:
    """
    Reference:

    - https://boto3.amazonaws.com/v1/documentation/api/latest/reference/services/sagemaker-a2i-runtime.html#AugmentedAIRuntime.Client.start_human_loop

    :return: the human in loop task ARN.
    """
    vprint(
        f"{emojis.play_or_pause} Start a Human Loop Task {human_loop_name!r}", verbose
    )
    flow_definition_name = parse_flow_definition_name_from_arn(flow_definition_arn)
    hil_console_url = get_hil_console_url(
        bsm.aws_region, flow_definition_name, human_loop_name
    )
    vprint(f"  You can preview HIL status at {hil_console_url}", verbose)
    input_data["hil_name"] = human_loop_name
    input_data["hil_console_url"] = hil_console_url
    kwargs = dict(
        HumanLoopName=human_loop_name,
        FlowDefinitionArn=flow_definition_arn,
        HumanLoopInput={
            "InputContent": json.dumps(input_data),
        },
    )
    if data_attributes is not None:
        kwargs["DataAttributes"] = data_attributes
    response = bsm.sagemaker_a2i_runtime_client.start_human_loop(**kwargs)
    hil_arn = response["HumanLoopArn"]
    return hil_arn


def get_human_loop_details(
    bsm: BotoSesManager,
    human_loop_name_or_arn: str,
) -> HumanLoop:
    """
    Get human loop details.
    """
    if human_loop_name_or_arn.startswith("arn:"):
        human_loop_name_or_arn = parse_hil_name_from_hil_arn(human_loop_name_or_arn)
    response = describe_human_loop(
        bsm=bsm,
        human_loop_name=human_loop_name_or_arn,
    )
    return HumanLoop(
        creation_time=response["CreationTime"],
        failure_reason=response.get("FailureReason"),
        failure_code=response.get("FailureCode"),
        human_loop_status=response["HumanLoopStatus"],
        human_loop_name=response["HumanLoopName"],
        human_loop_arn=response["HumanLoopArn"],
        flow_definition_arn=response["FlowDefinitionArn"],
        human_loop_output_s3uri=response["HumanLoopOutput"]["OutputS3Uri"],
        data=response,
    )


class HumanLoopIterProxy(IterProxy[HumanLoop]):
    pass


def _list_human_loops(
    bsm: BotoSesManager,
    flow_definition_arn: str,
    creation_time_after: T.Optional[datetime] = None,
    creation_time_before: T.Optional[datetime] = None,
    ascending: bool = False,
    max_items: T.Optional[int] = 1000,
    page_size: T.Optional[int] = 100,
) -> T.Iterator[HumanLoop]:
    paginator = bsm.sagemaker_a2i_runtime_client.get_paginator("list_human_loops")
    kwargs = dict(
        FlowDefinitionArn=flow_definition_arn,
    )
    if creation_time_after is not None:
        kwargs["CreationTimeAfter"] = creation_time_after
    if creation_time_before is not None:
        kwargs["CreationTimeBefore"] = creation_time_before
    if ascending:
        kwargs["SortOrder"] = "Ascending"
    else:
        kwargs["SortOrder"] = "Descending"
    kwargs["PaginationConfig"] = dict(
        MaxItems=max_items,
        PageSize=page_size,
    )
    response_iterator = paginator.paginate(**kwargs)
    parts = flow_definition_arn.split(":")
    aws_region = parts[3]
    aws_account_id = parts[4]
    for response in response_iterator:
        for data in response["HumanLoopSummaries"]:
            human_loop_name = data["HumanLoopName"]
            yield HumanLoop(
                creation_time=data["CreationTime"],
                failure_reason=data.get("FailureReason"),
                failure_code=None,
                human_loop_status=data["HumanLoopStatus"],
                human_loop_name=human_loop_name,
                human_loop_arn=f"arn:aws:sagemaker:{aws_region}:{aws_account_id}:human-loop/{human_loop_name}",
                flow_definition_arn=data["FlowDefinitionArn"],
                human_loop_output_s3uri=None,
                data=data,
            )


def list_human_loops(
    bsm: BotoSesManager,
    flow_definition_arn: str,
    creation_time_after: T.Optional[datetime] = None,
    creation_time_before: T.Optional[datetime] = None,
    ascending: bool = False,
    max_items: T.Optional[int] = 1000,
    page_size: T.Optional[int] = 100,
) -> HumanLoopIterProxy:
    """
    List human loops. You can then call :meth:`HumanLoop.get_details()` method
    to fetch more details.

    Reference:

    - https://boto3.amazonaws.com/v1/documentation/api/latest/reference/services/sagemaker-a2i-runtime.html#AugmentedAIRuntime.Client.list_human_loops
    """
    return HumanLoopIterProxy(
        iterable=_list_human_loops(
            bsm=bsm,
            flow_definition_arn=flow_definition_arn,
            creation_time_after=creation_time_after,
            creation_time_before=creation_time_before,
            ascending=ascending,
            max_items=max_items,
            page_size=page_size,
        )
    )



# -*- coding: utf-8 -*-

import typing as T
import uuid
import dataclasses
from datetime import datetime

from boto_session_manager import BotoSesManager

from .better_boto.human_task_ui import deploy_hil_task_template
from .better_boto.human_task_ui import remove_hil_task_template
from .better_boto.flow_definition import FlowDefinition
from .better_boto.flow_definition import is_flow_definition_exists
from .better_boto.flow_definition import deploy_flow_definition
from .better_boto.flow_definition import get_flow_definition_arn
from .better_boto.flow_definition import remove_flow_definition
from .better_boto.human_loop import parse_team_name_from_private_team_arn
from .better_boto.human_loop import get_workspace_signin_url
from .better_boto.human_loop import start_human_loop
from .better_boto.human_loop import list_human_loops


@dataclasses.dataclass
class Play:
    aws_account_id: T.Optional[str] = dataclasses.field(default=None)
    aws_region: T.Optional[str] = dataclasses.field(default=None)
    task_template_name: T.Optional[str] = dataclasses.field(default=None)
    task_template_content: T.Optional[str] = dataclasses.field(default=None)
    flow_definition_name: T.Optional[str] = dataclasses.field(default=None)
    flow_execution_role_arn: T.Optional[str] = dataclasses.field(default=None)
    output_bucket: T.Optional[str] = dataclasses.field(default=None)
    output_key: T.Optional[str] = dataclasses.field(default=None)
    private_team_arn: T.Optional[str] = dataclasses.field(default=None)

    @property
    def flow_definition_arn(self) -> str:
        return get_flow_definition_arn(
            aws_account_id=self.aws_account_id,
            aws_region=self.aws_region,
            flow_definition_name=self.flow_definition_name,
        )

    def get_flow_definition(
        self,
        bsm: BotoSesManager,
    ) -> T.Optional[FlowDefinition]:
        _, flow_def = is_flow_definition_exists(
            bsm=bsm,
            flow_definition_name=self.flow_definition_name,
        )
        return flow_def

    def deploy_hil_task_template(
        self,
        bsm: BotoSesManager,
        tags: T.Optional[T.Dict[str, str]] = None,
        verbose: bool = True,
    ):
        return deploy_hil_task_template(
            bsm=bsm,
            task_template_name=self.task_template_name,
            task_template_content=self.task_template_content,
            tags=tags,
            verbose=verbose,
        )

    def deploy_flow_definition(
        self,
        bsm: BotoSesManager,
        task_description: str,
        task_count: int,
        task_availability_life_time_in_seconds: T.Optional[int] = None,
        task_time_limit_in_seconds: T.Optional[int] = None,
        tags: T.Optional[T.Dict[str, str]] = None,
        wait: bool = True,
        wait_delay: int = 3,
        wait_timeout: int = 30,
        verbose: bool = True,
    ):
        return deploy_flow_definition(
            bsm=bsm,
            flow_definition_name=self.flow_definition_name,
            flow_execution_role_arn=self.flow_execution_role_arn,
            labeling_team_arn=self.private_team_arn,
            output_bucket=self.output_bucket,
            output_key=self.output_key,
            task_template_name=self.task_template_name,
            task_description=task_description,
            task_count=task_count,
            task_availability_life_time_in_seconds=task_availability_life_time_in_seconds,
            task_time_limit_in_seconds=task_time_limit_in_seconds,
            tags=tags,
            wait=wait,
            wait_delay=wait_delay,
            wait_timeout=wait_timeout,
            verbose=verbose,
        )

    def get_workspace_signin_url(
        self,
        bsm: BotoSesManager,
    ):
        return get_workspace_signin_url(
            bsm=bsm,
            work_team_name=parse_team_name_from_private_team_arn(
                arn=self.private_team_arn
            ),
        )

    def start_human_loop(
        self,
        bsm: BotoSesManager,
        input_data: dict,
        human_loop_name: T.Optional[str] = None,
        data_attributes: T.Optional[dict] = None,
        verbose: bool = True,
    ):
        if human_loop_name is None:
            human_loop_name = str(uuid.uuid4())
        return start_human_loop(
            bsm=bsm,
            human_loop_name=human_loop_name,
            flow_definition_arn=self.flow_definition_arn,
            input_data=input_data,
            data_attributes=data_attributes,
            verbose=verbose,
        )

    def remove_flow_definition(
        self,
        bsm: BotoSesManager,
        wait: bool = True,
        wait_timeout: int = 30,
        verbose: bool = True,
    ):
        return remove_flow_definition(
            bsm=bsm,
            flow_definition_name=self.flow_definition_name,
            wait=wait,
            wait_timeout=wait_timeout,
            verbose=verbose,
        )

    def remove_hil_task_template(
        self,
        bsm: BotoSesManager,
        verbose: bool = True,
    ):
        return remove_hil_task_template(
            bsm=bsm,
            task_template_name=self.task_template_name,
            verbose=verbose,
        )

    def list_human_loops(
        self,
        bsm: BotoSesManager,
        creation_time_after: T.Optional[datetime] = None,
        creation_time_before: T.Optional[datetime] = None,
        ascending: bool = False,
        max_items: T.Optional[int] = 1000,
        page_size: T.Optional[int] = 100,
    ):
        return list_human_loops(
            bsm=bsm,
            flow_definition_arn=self.flow_definition_arn,
            creation_time_after=creation_time_after,
            creation_time_before=creation_time_before,
            ascending=ascending,
            max_items=max_items,
            page_size=page_size,
        )
